/**
 * Requires: utils.js
 */

/**
 * Converts all the node indices (`id`, `source` and `target`) for each item in the given list.
 * @param {Object | Array} itemList Object or list of objects for conversion
 */
 function _indicesToString(itemList) {
    attributesToString(itemList, "id", "source", "target");
}

/**
 * Gets the edge identifier with node `u` and `v` as two ends. 
 * 
 * The edge identifier is represented as `(u', v')`, where `u' <= v'` (by lexicographical order) is always satisfied.
 * 
 * @param {*} u Identifier of node `u`
 * @param {*} v Identifier of node `v`
 * @returns The identifier of the corresponding edge (u, v) with the format above. 
 */
 function getEdgeId(u, v) {
    return `(${u <= v ? u : v}, ${u <= v ? v : u})`;
}

/**
 * Reads the configuration parameters from a JSON file.
 * @param {string} configPath Local path or remote URL of the config file
 * @returns An object that contains all the configurations.
 * @see `config-guide.md` for details of configuration file contents. 
 */
async function readConfig(configPath) {
    const conf = await fetch(configPath).then((t) => t.json());
    
    // Required attributes:
    //  (1) nodeStyles and edgeStyles
    checkAttributesAs(conf, ['nodeStyles', 'edgeStyles'], Object);
    //  (2) (node|edge)Styles contain 5 attributes: default (no message), Ca, Ca+, Cr, Cr-
    checkAttributesAs([conf['nodeStyles'], conf['edgeStyles']], ['default', 'Ca', 'Ca+', 'Cr', 'Cr-'], Object);
    
    ['Ca', 'Ca+', 'Cr', 'Cr-'].forEach((s) => {
        // (3) Static node & animation node
        checkAttributesAs(conf['nodeStyles'][s], ['default', 'animation'], Object);
    });

    // Optional attributes:
    //  (1) nAnimationNodes: 10 by default
    checkAttributes(conf, 'nAnimationNodes', (obj, attr) => {
        obj[attr] = 10; 
    });
    //  (2) boosted or source node styles
    function addNodeStyle(obj, name, defaultName = "default") {
        if (obj[name] == undefined) {
            obj[name] = obj[defaultName];
        }
    }
    addNodeStyle(conf.nodeStyles, "boosted");
    ['Ca', 'Ca+', 'Cr', 'Cr-'].forEach((s) => {
        ['boosted', 'source'].forEach((t) => addNodeStyle(conf.nodeStyles[s], t));
    });

    return Object.freeze(conf); 
}

/**
 * Reads the graph from a file.
 * 
 * The graph is given as a JSON object file including the following contents:
 * ```
 * "nodes": [
 *   { "id": 0 },   // Each node has an index 
 *   { "id": 1 },
 * ]    // A list of all the nodes
 * "edges": [
 *   { "source": 0, "target": 1 },      // Each edge (source, target) is bidirectional
 *   { "source": 1, "target": 2 }   
 * ],   // A list of all the bidirectional edges
 * ```
 * The graph shall be **bidirectional** and satisfy the following criteria:
 *   * Connected: multiple connected components are disallowed;
 *   * Unique node: duplicated nodes (sharing the same id) are disallowed;
 *   * Unique edge: duplicated edges are disallowed (`{source: u, target: v}` and `{source: v, target: u}` are regarded the same);
 *   * Size constraints: Due to performance limit of rendering, too large graphs are not suggested.
 * 
 * The graph object returned contains the following attributes:
 *   * `nodes`: A list of nodes, with id converted to string;
 *   * `edges`: A list of bidirectional edges, with node ids converted to string;
 *   * `n`: Number of nodes, equivalent to nodes.length;
 *   * `m`: Number of edges, equivalent to edges.length;
 *   * `nodeMap`: A map(string => Object) from index to the corresponding node object.
 * 
 * Each node has the following attributes appended (if not provided in the input JSON file):
 *   * `degree`: Degree of the node
 * 
 * Each link has the following attributes appended (if not provided in the input JSON file):
 *   * `id`: Identifier of the edge
 * 
 * **Warning: If the additional attributes are provided via input, it's your own duty to 
 * ensure their correctness.**
 * 
 * @param {string} path Local path or remote URL of the graph file
 * @returns The graph object
 * @see getEdgeId for details of format of edge identifier.
 */
async function readGraph(path) {
    const graph = await fetch(path).then((t) => t.json());

    // 'id' attribute should be transformed to String
    _indicesToString(graph.nodes);
    _indicesToString(graph.edges);

    // Number of nodes in the graph
    graph.n = graph.nodes.length;
    // Number of edges in the graph 
    graph.m = graph.edges.length;
    console.log(`${path}: |V| = ${graph.n}, |E| = ${graph.m}`);

    // Builds the node map: id => node
    graph.nodeMap = new Map();
    graph.nodes.forEach((node) => {
        graph.nodeMap.set(node.id, node); 
    });

    // Degree is calculated if not provided by input
    if (graph.nodes.length > 0 && graph.nodes[0].degree == undefined) {
        graph.nodes.forEach((node) => {
            // Initializes degrees to 0
            node.degree = 0; 
        });
        graph.edges.forEach((edge) => {
            for (const id of [edge.source, edge.target]) {
                graph.nodeMap.get(id).degree += 1;
            }
        });
    }
    // Identifiers of each edge is assigned if not provided by input
    if (graph.edges.length > 0 && graph.edges[0].id == undefined) {
        graph.edges.forEach((edge) => {
            const u = edge.source;
            const v = edge.target;
            edge.id = getEdgeId(u, v);
        })
    }

    return graph;
}

/**
 * Creates an object as a collection of all the events, including the following attributes:
 *   * `label`: Label of current group of data;
 *   * `nRounds`: Total number of rounds (such that each timestamp falls in range [0, nRounds-1]);
 *   * `boostedNodes': List of boosted nodes; 
 *   * `sourceEvents`: A 2D-array of all the events of source events;
 *   * `propagationEvents`: A 2D-array of all the events of propagation events;
 *   * `maxNPropagations`: Maximum number of propagation events in each timestamp.
 * 
 * For each timestamp *t*, `sourceEvents[t]` and `propagationEvents[t]` contain all the corresponding events
 * at timestamp *t*. 
 * 
 * `sourceEvents` and `propagationEvents` are initialized as an array of length `nRounds`, with each element inside
 * initialized as en empty array. 
 * 
 * @param {String} label Label of current group of data;
 * @param {Number} nRounds Number of rounds;
 * @param {Array<String>} boostedNodes List of boosted nodes, empty by default; 
 * @param {Number} maxNPropagations Maximum number of propagation events in a round. `NaN` as default if not provided.
 * @returns An event collection object. 
 * @see getEventListsObject
 */
function createEventObject(label, nRounds, boostedNodes = [], maxNPropagations = NaN) {
    const res = {
        label: label,
        nRounds: nRounds,
        boostedNodes: boostedNodes,
        sourceEvents: d3.range(nRounds).map((_) => []),
        propagationEvents: d3.range(nRounds).map((_) => []),
        maxNPropagations: maxNPropagations
    };
    return res; 
}

/**
 * Reads the data about propagation information from a file.
 * 
 * The data are given as a JSON list which may include the following types of objects:
 * ```
 * {
 *   "boostedNodes": [a, b, c],
 *   "sourceEvents": [
 *     { 
 *       "timestamp": t,
 *       "id": u,
 *       "type": "Ca" | "Ca+" | "Cr" | "Cr-"
 *     }
 *   ],
 *   "propagationEvents": [
 *     {
 *       "timestamp": t,
 *       "source": u,
 *       "target": v,
 *       "edgeType": "Ca" | "Ca+" | "Cr" | "Cr-",
 *       "targetType": "Ca" | "Ca+" | "Cr" | "Cr-"
 *     }
 *   ]
 * }
 * ```
 * The result object is initialized via `createEventObject` function. 
 * 
 * Each item in the event lists has all the node indices converted to string. 
 * 
 * @param {string} path Local path or remote URL of the input file 
 * @param {string} label Identifier of current data file. 
 * @returns Object of the data collection, categorized by timestamp. 
 * @see createEventObject for details of each attribute in the object returned.
 */
async function readData(path, label) {
    const data = await fetch(path).then((t) => t.json());
    // Converts all the node indices to string
    [data.sourceEvents, data.propagationEvents].forEach((evtList) => _indicesToString(evtList));
    // Converts the boosted node list to string
    for (i in data.boostedNodes) {
        data.boostedNodes[i] = data.boostedNodes[i].toString(); 
    }
    // The number of rounds
    const nRounds = Math.max(
        d3.max(data["sourceEvents"].map((obj) => obj.timestamp)), 
        d3.max(data["propagationEvents"].map((obj) => obj.timestamp))
    ) + 1;

    const validTypeValues = ['Ca', 'Ca+', 'Cr', 'Cr-'];

    const errorOnTypeValue = (item, attr) => {
        const errInfo = `attribute "${attr}" is not one of ${validTypeValues}`;
        console.error(errInfo + " in object: ", item);
        throw errInfo; 
    };

    // Checks source event lists
    data.sourceEvents.forEach((item) => {
        checkAttributes(item, ["id", "type"]);
        if (!validTypeValues.includes(item.type)) {
            errorOnTypeValue(item, "type");
        }
    });
    // Checks propagation event lists
    data.propagationEvents.forEach((item) => {
        checkAttributes(item, ["source", "target", "edgeType", "targetType"]);
        ["edgeType", "targetType"].forEach((attr) => {
            if (!validTypeValues.includes(item[attr])) {
                errorOnTypeValue(item, attr); 
            }
        });
    });

    // maxNPropagations is initialized later.
    const res = createEventObject(label, nRounds, data.boostedNodes);

    // Categorizes each item by timestamp 
    ["sourceEvents", "propagationEvents"].forEach((evtListName) => {
        data[evtListName].forEach((item) => {
            res[evtListName][item.timestamp].push(item) 
        });
    });

    // Calculates maximum of propagation events in each timestamp
    res.maxNPropagations = d3.max(res.propagationEvents.map((arr) => arr.length));

    return res; 
}

/**
 * Reads multiple groups of data with given label and file path resp. 
 * 
 * The input `group` is an array of objects with the following attributes:
 *   * `label`: Label of current group of data;
 *   * `path`: Path of the JSON file (local file path or remote URL).
 * 
 * For each group of data, `nRounds`, along with the length of arrays `positiveAt` and `negativeAt`, 
 *  will be aligned to the maximum.
 * 
 * @param {Array<Object>} groups Information of data groups (see above).
 * @returns An object that maps each label to the corresponding data object. 
 * @see readData for details of processing with each data group.
 */
async function readDataGroup(groups) {
    // Starts all the tasks as future
    const futures = {};
    for (const g of groups) {
        futures[g.label] = readData(g.path, g.label); 
    }
    // Joins until all the futures finish
    const res = {};
    for (const [label, fut] of Object.entries(futures)) {
        res[label] = await fut; 
    }

    // Takes the maximum rounds. 
    const maxNRounds = d3.max(Object.values(res).map((data) => data.nRounds)); 
    Object.values(res).forEach((data) => {
        ['sourceEvents', 'propagationEvents'].forEach((evtListName) => {
            // Aligns array length with empty event lists
            d3.range(data.nRounds, maxNRounds).forEach((_) => {
                data[evtListName].push([]); 
            });
        });
        data.nRounds = maxNRounds; 
    });

    return res; 
}