/**
 * Checks whether the given attribute(s) exists in the given object(s).
 * 
 * The callback function `callback(obj, attr)` is invoked if the attribute does not exist in the object. 
 * If not provided by user, the default callback raises an error log in the console, and then
 * an exception will be thrown.
 * @param {Object | Array<Object>} obj The object, or the list of objects to be checked
 * @param {string | Array<string>} attr Attribute, or the list of the attributes to check
 * @param {function(Object, string)} callback The callback function if given attribute does not exist
 * @throws For the default callback, throws an error message if the attribute does not exist. 
 */
 function checkAttributes(obj, attr, callback = null) {
    const objList = obj instanceof Array ? obj : [obj];
    const attrList = attr instanceof Array ? attr : [attr];

    attrList.forEach((attr) => {
        objList.forEach((obj) => {
            if (obj[attr] == null) {
                if (callback == null) {
                    errInfo = `Missing config attribute '${attr}'`;
                    console.error(errInfo + " in object: ", obj);
                    throw errInfo; 
                } else {
                    callback(obj, attr);
                }
            }
        });
    });
}

/**
 * Checks whether the given attribute(s) exists and matches the specified type in the given object(s).
 * 
 * Each attribute is checked (if exists):
 *   * If type is a string **literal**, then `typeof obj[attr] == type` is checked;
 *   * Otherwise, `obj[attr] instanceof type` is checked.  
 * 
 * The callback function `callback(obj, attr)` is invoked if the attribute does not exist or has type mismatched in the object. 
 * If not provided by user, the default callback raises an error log in the console, and then
 * an exception will be thrown.
 * 
 * @param {Object | Array<Object>} obj The object, or the list of objects to be checked
 * @param {string | Array<string>} attr Attribute, or the list of the attributes to check
 * @param {*} type The expected type of each object
 * @param {*} callback The callback function if given attribute does not exist
 * @throws For the default callback, throws an error message if the attribute does not exist. 
 * @see checkAttributes
 */
function checkAttributesAs(obj, attr, type, callback = null) {
    const objList = obj instanceof Array ? obj : [obj];
    const attrList = attr instanceof Array ? attr : [attr];

    attrList.forEach((attr) => {
        objList.forEach((obj) => {
            if (obj[attr] == null) {
                if (callback == null) {
                    errInfo = `Missing attribute '${attr}'`;
                    console.error(errInfo + " in object: ", obj);
                    throw errInfo; 
                } else {
                    callback(obj, attr);
                }
            }
            else if (typeof type == 'string' && typeof obj[attr] != type) {
                if (callback == null) {
                    errInfo = `Wrong type of attribute '${attr}': Expected typeof '${type}' instead of '${typeof obj[attr]}'`;
                    console.error(errInfo + " in object: ", obj);
                    throw errInfo; 
                } else {
                    callback(obj, attr); 
                }
            }
            else if (typeof type == 'object' && !(obj[attr] instanceof type)) {
                if (callback == null) {
                    errInfo = `Wrong type of attribute '${attr}'`;
                    console.error(errInfo + " in object: ", obj);
                    throw errInfo; 
                } else {
                    callback(obj, attr); 
                }
            }
        });
    });
}

/**
* Converts all the given attributes (if exist) to string for each item in the given object list.
* @param {Object | Array} itemList Object, or list of objects for conversion
* @param {...string} attributes Attributes to convert.
*/
function attributesToString(itemList, ...attributes) {
    function doConvert(item) {
        attributes.forEach((attrName) => {
            if (item[attrName] != undefined) {
                item[attrName] = item[attrName].toString();
            }
        });
    }
    if (itemList instanceof Array) {
        itemList.forEach((item) => doConvert(item));
    } else {
        doConvert(item); 
    }
}

/**
 * Adds an attribute to `dest` as `dest[name] = value`. 
 * 
 * Checking is performed before assigning whether another attribute with the same identifier already exists.
 * A warning message to console is triggered and the old value is replaced 
 * if duplicated attribute identifiers are detected. 
 * @param {Object} dest The object where the new attribute is appended
 * @param {String} name Identifier of the new attribute
 * @param {any} value Value of the new attribute
 */
function createAttribute(dest, name, value) {
    if (dest[name] != undefined) {
        console.warn(`WARNING: rewriting attribute '${name}' that already exists.`);
    }
    dest[name] = value; 
}